# -*- coding: utf-8 -*-
"""
Created on Sat Apr 11 12:20:32 2020

@author: Germán Riaño (griano@germanriano.com)
"""


import logging
import math

import numpy as np
import pandas as pd
from scipy import integrate, interpolate
from tqdm import tqdm

from epistoch.utils.stats import loss_function
from epistoch.utils.utils import compute_integral, get_total_infected

EPS = 1e-5


# The SIR model differential equations.
def _sir_deriv(t, y, beta, gam):
    S, I = y
    dSdt = -beta * S * I
    dIdt = beta * S * I - gam * I
    return dSdt, dIdt


def sir_classical(name, population, reproductive_factor, infectious_period_mean, num_days, I0=1.0, S0=None, times=None):
    """
    Solves a classical SIR model.

    Parameters
    ----------
    name: string
        Model name
    population: float
        total population size
    reproductive_factor: float
        basic reproductive factor(R0)
    infectious_period_mean: float
        expected value of Infectious Period Time
    I0: float
        Initial infectious population
    S0: float
        Initial susceptible population (optinal, defaults to all but I0)
    num_days: int
        number of days to run
    times: array of float
        times where the functions should be reported. Defaults to
        ``np.linspace(0, num_days, num_days + 1)``


    Returns
    -------
    dict
        Dictionary with fields:
            - name: model name
            - population: Total population
            - total_infected
            - data: data Frame with columns
                - S : Susceptible,
                - I : Infectious (Dying or recovering),
                - R : Removed (recovered + deaths),
    """
    # Note: Code based on
    # https://scipython.com/book/chapter-8-scipy/additional-examples/the-sir-epidemic-model/
    # Total population, N.
    N = population
    # Initial number of infected and recovered individuals, I0 and R0.
    I0 = I0 / N
    S0 = S0 / N if S0 is not None else 1 - I0
    # Note that R0 = 1 - I0 - S0

    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1.0 / infectious_period_mean
    beta = reproductive_factor * gam

    # A grid of time points (in days)
    times = np.linspace(0, num_days, num_days + 1) if times is None else times

    # Initial conditions vector
    y0 = [S0, I0]
    # Integrate the SIR equations over the time grid, t.
    ret = integrate.odeint(_sir_deriv, y0, times, args=(beta, gam), tfirst=True)
    S, I = ret.T
    # De-normalize
    S = S * N
    I = I * N
    R = N - S - I
    result = dict()
    result["data"] = pd.DataFrame(data={"Day": times, "S": S, "I": I, "R": R}).set_index("Day")
    result["total_infected"] = get_total_infected(reproductive_factor)
    result["population"] = N
    result["name"] = name
    return result


def sir_g(
    name,
    population,
    reproductive_factor,
    infectious_time_distribution,
    num_days,
    I0=1.0,
    S0=None,
    num_periods=None,
    method="loss",
    logger=None,
):
    """
    Solves a SIR-G model.

    Parameters
    ----------
    name: string
        Model name
    population: float
        total population size
    reproductive_factor: float
        basic reproductive factor(R0)
    infectious_time_distribution: scipy.stats.rv_continuous
        expected value of Infectious Period Time
    num_days: int
        number of days to run
    I0: float
        Initial infectious population
    S0: float
        Initial susceptible population (optional, defaults to all but I0)
    num_periods:int
        Number of periods to use for computations. Higher number will lead ot more precise computation.
    method: string
        Method used for the internal integral
    logger
        Logger object

    Returns
    -------
    dict
        Dictionary with fields:
            - name: model name
            - population: Total population
            - total_infected
            - data: data Frame with columns
                - S : Susceptible,
                - I : Infectious (Dying or recovering),
                - R : Removed (recovered + deaths),
    """
    if logger is None:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)
    # Total population, N.
    N = population
    dist = infectious_time_distribution

    I0 = I0 / N
    S0 = S0 / N if S0 is not None else 1 - I0
    # Notice that R0 = 1 - I0 - S0

    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1 / dist.mean()
    beta = reproductive_factor * gam
    num_periods = 10 * num_days if num_periods is None else num_periods
    delta = num_days / num_periods
    times = np.linspace(start=0.0, stop=num_days, num=num_periods + 1)
    logging.info("Computing survival function")
    survival = dist.sf(times)
    logging.info("Computing PDF function")
    pdfs = dist.pdf(times)
    if pdfs[0] == math.inf:
        pdfs[0] = dist.pdf(dist.ppf(0.01))
    survival = dist.sf(times)
    if survival[0] < 1.0 - EPS:
        logging.warning(f"Survival function does not start in 1:  {survival}")
    logging.info("Computing Loss1 function")
    loss_fun = loss_function(dist)
    loss1 = loss_fun(times)

    S = np.zeros(num_periods + 1)
    # S[:] = np.nan
    I = np.zeros(num_periods + 1)
    #    I[:] = np.nan
    S[0] = S0
    I[0] = I0
    logging.info(f"Computing SIR-G model with {num_periods} periods.")
    for n in tqdm(range(0, num_periods)):
        S[n + 1] = S[n] - delta * beta * S[n] * I[n]
        integral = compute_integral(n, delta, S, I, times, survival, pdfs, loss1, dist, method=method)
        # Expected case for exponential variables
        integral_expon = (gam / beta) * (I[n] - I0 * survival[n])
        logger.debug(f"n={n}, t={n*delta}, integral = {integral}, integral_expon = {integral_expon}")
        I[n + 1] = I[n] + delta * (beta * (S[n] * I[n] - integral) - gam * I0 * survival[n])

    # Report one point per day, and de-normalize
    days = np.linspace(0, num_days, num_days + 1)
    S = N * interpolate.interp1d(times, S)(days)
    I = N * interpolate.interp1d(times, I)(days)
    R = N - S - I
    result = dict()
    result["name"] = name
    result["population"] = population
    result["data"] = pd.DataFrame(data={"Day": days, "S": S, "I": I, "R": R}).set_index("Day")
    result["total_infected"] = get_total_infected(reproductive_factor)
    return result



import numpy as np
from scipy import integrate, interpolate, optimize


def compute_integral(n, delta, S, I, times, survival, pdfs, loss1, dist, method="loss"):
    r"""

    Compute the integral needed for the integro-differential model.

    In other words, computes

    .. math:: \int_0^t g(t-x) I(x)S(x) dx \quad  \text{ for } t = n*\delta

    Parameters
    ----------
    n : integer
        upper limit for integral.
    delta : float
        interval size
    S : array
        Susceptible
    I : array
        Infected
    times: array
        times at which the arraya are evaluated
    survival : array of float
        array :math:`G_k \equiv P\{ T > k\delta\}`.
    pdfs: array of float
        array :math:`g(k\delta)`.
    loss1 array float
        :math:`L_k\equiv L(k\delta)`
    dist: rv_continuous
        Object that represents the distribution
    method: string
        One from "loss", "simpson" or "interpolate"
    Returns
    -------
    Value of the integral

    """
    if n == 0:
        return 0.0
    if method == "loss":
        IS = np.zeros_like(survival)
        # next line equivalent to: IS[: n + 1] = np.array([I[n - k] * S[n - k] for k in range(n + 1)])
        IS[: n + 1] = np.flipud(I[: n + 1] * S[: n + 1])
        slopes = np.diff(IS, append=0.0) / delta  # m1, m2, ...
        delta_slopes = np.diff(slopes, prepend=0.0)
        return IS[0] + sum(delta_slopes * loss1)

    if method == "simpson":
        integral_points = [pdfs[n - k] * S[k] * I[k] for k in range(0, n + 1)]
        return integrate.simps(integral_points, dx=delta)
    if method == "interpolate":
        t = n * delta
        interpolator = interpolate.interp1d(times[: n + 1], S[: n + 1] * I[: n + 1])

        def _integrand(tau):
            return dist.pdf(t - tau) * interpolator(tau)

        return integrate.quad(_integrand, 0, t)[0]


def get_total_infected(reproductive_factor, normalized_s0=1):
    r"""
    Estimate the total number of infected for a given reproductive factor.

    Find the value z that solves

    .. math:: 1-z = S_0 * e^{{\mathcal R_0} z},

    where :math:`\mathcal R_0` is the ``reproductive_factor``, and :math:`S_0` is the (normalized) initial
    population ``normalized_s0``.

    Parameters
    ----------
    reproductive_factor : float
        Basic reproductive factor.
    normalized_s0 : float, optional
        Initial fraction of population that is infected. The default is 1.

    Returns
    -------
    float
        The stimated fraction of total infected.

    """
    if reproductive_factor < 1:
        return 0.0
    else:
        fun = lambda x: 1 - x - normalized_s0 * np.exp(-reproductive_factor * x)
        result = optimize.root_scalar(fun, bracket=[0.0001, 1])
        return result.root


def _compute_array_error(name, x1, x2, N, do_print=True):
    # Convenience method to compute the error difference in tow array
    error = np.abs(x1 - x2) / N
    if do_print:
        print(f"{name}: max error = {np.max(error):.2}, avg error = {np.mean(error):.2}")
    return np.max(error)


def _compute_error(model1, model2, N, do_print=True):
    # Convenience methods to compute errors between two models
    if model1["population"] != model2["population"]:
        raise ValueError("We can only compare models with the same population")
    N = model1["population"]
    error_i = _compute_array_error("I", model1["data"].I, model2["data"].I, N, do_print)
    error_s = _compute_array_error("S", model1["data"].S, model2["data"].S, N, do_print)
    return 0.5 * (error_i + error_s)


def print_error(model1, model2):
    """
    Print error difference between two models

    Parameters
    ----------
    model1
        Result of a SIR, SIR_G model
    model2
        Result of a SIR, SIR_G model

    Returns
    -------

    """
    return _compute_error(model1, model2, True)


def report_summary(result):
    """
    Report a summary for a model.

    Parameters
    ----------
    result:
        Result from SIR like model
    """
    model = result["data"]
    name = result["name"]
    N = result["population"]
    n = len(model) - 1
    print(f"Model {name} Summary")
    print(f"  Total Infected People: {int(model.R[n]):,d} ({model.R[n]/N:.2%})")
    print(f"  Infection Peak: {int(np.max(model.I)):,d} ({np.max(model.I)/N:.2%})")
    print(f"  Peak Day: {int(np.argmax(model.I)):,d}")
    print(f"  Theoretical Total Infected: {result['total_infected']:.2%}")



"""Top-level package for EpiStoch."""

__author__ = """Germán Riaño"""
__email__ = "griano@germanriano.com"
__version__ = "0.1.13"

from .seird_ph import *
from .sir_g import *
from .sir_phg import *



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 12 16:50:30 2020

@author: Germán Riaño
"""

import logging
import math

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy import integrate, interpolate

import epistoch.utils as utils
from epistoch import compute_integral, get_total_infected, sir_classical
from epistoch.sir_g import EPS, sir_g
from epistoch.utils.plotting import plot_sir
from epistoch.utils.stats import loss_function
from epistoch.utils.utils import _compute_array_error, print_error


def _deriv2(
    # Derivatives for second step. S and I are fixed.
    t,
    y,
    beta,
    gam,
    I0,
    times,
    delta,
    S_guess,
    I_guess,
    survival,
    pdfs,
    loss1,
    dist,
    method,
):
    # This function does not depend on y, but it depends on t
    # We buid piece-wise linear functions
    if t < min(times) or t > max(times):
        return 0.0, 0.0

    S = interpolate.interp1d(times, S_guess, fill_value="extrapolate")
    I = interpolate.interp1d(times, I_guess, fill_value="extrapolate")
    dSdt = -beta * S(t) * I(t)
    n = int(np.floor(t / delta))
    integral = compute_integral(n, delta, S_guess, I_guess, times, survival, pdfs, loss1, dist, method)
    dIdt = -dSdt - beta * integral - gam * I0 * survival[n]

    return dSdt, dIdt


def sir_g2(
    name,
    population,
    reproductive_factor,
    infectious_time_distribution,
    num_days,
    I0=1.0,
    S0=None,
    num_periods=None,
    method="loss",
    logger=None,
):
    if logger is None:
        logger = logging.getLogger(__name__)
        logger.setLevel(logging.INFO)

    # Total population, N.
    N = population
    dist = infectious_time_distribution
    # Normailize to imporve numerical stability
    I0 = I0 / N
    S0 = S0 / N if S0 is not None else 1 - I0
    # Notice that R0 = 1 - I0 - S0

    # Contact rate, beta, and mean recovery rate, gam, (in 1/days).
    gam = 1 / dist.mean()
    beta = reproductive_factor * gam
    num_periods = 10 * num_days if num_periods is None else num_periods
    delta = num_days / num_periods
    times = np.linspace(start=0.0, stop=num_days, num=num_periods + 1)
    survival = dist.sf(times)
    pdfs = dist.pdf(times)
    if pdfs[0] == math.inf:
        pdfs[0] = dist.pdf(dist.ppf(0.01))
    survival = dist.sf(times)
    if survival[0] < 1.0 - EPS:
        logging.warning(f"Survival function does not start in 1:  {survival}")
    loss_fun = loss_function(dist)
    loss1 = loss_fun(times)

    S = np.empty(num_periods + 1)
    S[:] = np.nan
    I = np.empty(num_periods + 1)
    I[:] = np.nan
    S[0] = S0
    I[0] = I0

    # Initial conditions vector
    y0 = [S0, I0]

    max_iterations = 100
    logger.info("Computing classical SIR")
    sir_class = sir_classical(
        name="Classic",
        population=1,
        reproductive_factor=reproductive_factor,
        infectious_period_mean=dist.mean(),
        num_days=num_days,
        I0=I0,
        times=times,
    )
    data = sir_class["data"]
    old_y = np.asarray((np.asarray(data.S), np.asarray(data.I)))
    # old_y = np.asarray((.5*np.ones_like (times), .5* np.ones_like (times)))
    # old_y = np.asarray((np.zeros_like (times), np.zeros_like (times)))
    max_diff = 1e-6
    alpha = 0.9
    for iteration in range(max_iterations):
        S_guess, I_guess = old_y
        # Integrate again, with the integral, fixing guess
        args = (beta, gam, I0, times, delta, S_guess, I_guess, survival, pdfs, loss1, dist, "loss")
        logger.debug(f"Staring iteration {iteration:d}")
        ret2 = integrate.odeint(func=_deriv2, y0=y0, t=times, args=args, tfirst=True)
        y = ret2.T
        S, I = y
        diff = _compute_array_error("I", I_guess, I, N, do_print=False)
        if logger.isEnabledFor(logging.DEBUG):
            pd.DataFrame(data={"Times": times, "S": S, "I": I}).set_index("Times").plot()
            plt.show()
        logger.debug(f"Iteration {iteration:d}: {diff:.4%}")
        if diff < max_diff:
            break
        old_y = alpha * y + (1 - alpha) * old_y

    # Report one point per day, and de-normalize
    days = np.linspace(0, num_days, num_days + 1)
    S = N * interpolate.interp1d(times, S)(days)
    I = N * interpolate.interp1d(times, I)(days)
    R = N - S - I
    result = dict()
    result["data"] = pd.DataFrame(data={"Day": days, "S": S, "I": I, "R": R}).set_index("Day")
    result["total_infected"] = get_total_infected(reproductive_factor)
    result["population"] = N
    result["name"] = name
    return result


def compare_sir_g2(dist, num_periods=2000, logger=None):
    N = 1000
    reproductive_factor = 2.2
    num_days = 160
    model0 = sir_classical(
        "SIR_CL",
        population=N,
        reproductive_factor=reproductive_factor,
        infectious_period_mean=dist.mean(),
        num_days=num_days,
    )
    model1 = sir_g(
        "SIR_G",
        population=N,
        reproductive_factor=reproductive_factor,
        infectious_time_distribution=dist,
        num_days=num_days,
        num_periods=num_periods,
        logger=None,
    )
    model2 = sir_g2(
        "SIR_G2",
        population=N,
        reproductive_factor=reproductive_factor,
        infectious_time_distribution=dist,
        num_days=num_days,
        num_periods=num_periods,
        logger=logger,
    )
    print_error(model1, model2)
    fig = plot_sir(model1)
    plot_sir(model2, fig=fig, linestyle="--")
    plot_sir(model0, fig=fig, linestyle=":")
    plt.show()


def test_sir_g2(dist=utils.stats.get_gamma(10, 2), num_periods=2000, logger=None):
    compare_sir_g2(dist, num_periods, logger)


if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    test_sir_g2(dist=utils.stats.get_gamma(10, 20), num_periods=2000, logger=logger)
